//The contents of this file is public domain. No warranty implied. Use at your own risk.
#ifdef REFLECTION
#undef REFLECTION
#endif
#define REFLECTION %0
#ifdef SPHERESIZE
#undef SPHERESIZE
#endif
#define SPHERESIZE %1
#ifdef COLOR
#undef COLOR
#endif
#define COLOR %2


kernel void distmap%0%1%2(global half* restrict spheres
                    ,global half* restrict bounding_spheres
                    ,global unsigned int* restrict bounding_ref
#if (COLOR==1) || (REFLECTION==1)
					,global half* restrict sphere_colors
#endif
                    ,global float4* restrict ray_origin
                    ,float4 big_bound
                    ,int bounding_start
                    ,int bounding_end
                    ,unsigned int start
                    ,unsigned int end
                    ,global float4* restrict dir
                    ,global RaytraceInfo* restrict d
#if (COLOR==0) || (REFLECTION==0)
					,float4 color_reflect
#endif
#if (SPHERESIZE==0)
					,float sphere_size
#endif
					)
{
    int x = get_global_id(0);
    RaytraceInfo rtmp = d[x];
    if(rtmp.state != 2){
		int id = -1;
		int change = 0;
        float4 ddir = dir[x];
        float4 origin = ray_origin[x];
        float sq_bbradius = big_bound.w * big_bound.w;
        float3 bbp_a = big_bound.xyz - origin.xyz;
        float bbp_an = dot(bbp_a, ddir.xyz);
        float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));
        if(sq_bbndist <= sq_bbradius){
			for(int b = bounding_start; b <= bounding_end; b++){
				float4 bsphere = vloada_half4(b, bounding_spheres);
				float sq_bradius = bsphere.w * bsphere.w;	
				//float sq_bradius = bsphere.w; //if there are cracks at the edges of the bounding spheres
				float3 bp_a = bsphere.xyz - origin.xyz;
				float bp_an = dot(bp_a, ddir.xyz);
				float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
				if(sq_bndist <= sq_bradius){
					if ((bp_an - bsphere.w) < rtmp.closest_distance){
						float bdist_exact = bp_an - native_sqrt(sq_bradius - sq_bndist);
						if (bdist_exact < rtmp.closest_distance){
							unsigned int startix = (b == 0) ? 0 : bounding_ref[b-1];
							startix = max(start,startix);
							unsigned int endix = min(end, bounding_ref[b]);
							for(unsigned int i = startix; i < endix; i++){
								float4 sphere;
#if (SPHERESIZE==1)
								sphere = vloada_half4(i, spheres);
#else
								sphere.x = vloada_half(i*3, spheres);
								sphere.y = vloada_half(i*3+1, spheres);
								sphere.z = vloada_half(i*3+2, spheres);
								sphere.w = sphere_size;
#endif
								sphere.xyz += bsphere.xyz;
								float sq_radius = sphere.w * sphere.w;
								float3 p_a = sphere.xyz - origin.xyz;
								float p_an = dot(p_a, ddir.xyz);
								float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
								if(sq_ndist <= sq_radius){
									if ((p_an - sphere.w) < rtmp.closest_distance){
										float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
										if (dist_exact > 0.0f && dist_exact < rtmp.closest_distance){
											rtmp.state = change = 1;
											rtmp.closest_distance = dist_exact;
											rtmp.sphere_hit = sphere;//.xyz;
											id = i;
										}
									}
								}
							}
						}
					}
				}
			}
			if(change){
				if(id >= 0) {
#if (COLOR==0) && (REFLECTION==0)
					rtmp.sphere_color = color_reflect;
#elif (COLOR==1) && (REFLECTION==1)
					rtmp.sphere_color = vloada_half4(id, sphere_colors);
#elif (COLOR==1) && (REFLECTION==0)
					rtmp.sphere_color.x = vloada_half(id*3, sphere_colors);
					rtmp.sphere_color.y = vloada_half(id*3+1, sphere_colors);
					rtmp.sphere_color.z = vloada_half(id*3+2, sphere_colors);
					rtmp.sphere_color.w = color_reflect.w;
#else
					rtmp.sphere_color.xyz = color_reflect.xyz;
					rtmp.sphere_color.w = vloada_half(id, sphere_colors);
#endif
				}
				d[x] = rtmp;
			}
        } //if(sq_bbndist <= sq_bbradius)
    } //if(rtmp.state != 2)
}

