//The contents of this file is public domain. No warranty implied. Use at your own risk.
typedef struct tag_RaytraceInfo
{
    int state;
	float closest_distance, mirror_factor;
	unsigned int shadow;
	float4 sphere_hit;
	float4 sphere_color;
} RaytraceInfo;

float sq_point(float3 p, float3 c)
{
    return ((p.x - c.x) * (p.x - c.x)) + ((p.y - c.y) * (p.y - c.y)) + ((p.z - c.z) * (p.z - c.z));
}

float3 reflect(float3 I, float3 N)
{
	return I - N * dot(N, I) * 2.0f;
}

kernel void distmap(global float4* restrict spheres, 
                    global float4* restrict bounding_spheres, 
                    global unsigned int* restrict bounding_ref, 
					global float4* restrict sphere_colors,
                    global float4* restrict ray_origin,
                    float4 big_bound,
                    int bounding_start,
                    int bounding_end,
                    unsigned int start,
                    unsigned int end, 
                    global float4* restrict dir, 
                    global RaytraceInfo* restrict d)
{
    int x = get_global_id(0);
    RaytraceInfo rtmp = d[x];
    if(rtmp.state != 2){
		int id = -1;
		int change = 0;
        float4 ddir = dir[x];
        float4 origin = ray_origin[x];
        float sq_bbradius = big_bound.w * big_bound.w;
        float3 bbp_a = big_bound.xyz - origin.xyz;
        float bbp_an = dot(bbp_a, ddir.xyz);
        float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));
        if(sq_bbndist <= sq_bbradius){
			for(int b = bounding_start; b <= bounding_end; b++){
				float4 bsphere = bounding_spheres[b];
				float sq_bradius = bsphere.w * bsphere.w;
				float3 bp_a = bsphere.xyz - origin.xyz;
				float bp_an = dot(bp_a, ddir.xyz);
				float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
				if(sq_bndist <= sq_bradius){
					if ((bp_an - bsphere.w) < rtmp.closest_distance){
						float bdist_exact = bp_an - native_sqrt(sq_bradius - sq_bndist);
						if (bdist_exact < rtmp.closest_distance){
							unsigned int startix = (b == 0) ? 0 : bounding_ref[b-1];
							startix = max(start,startix);
							unsigned int endix = min(end, bounding_ref[b]);
							for(unsigned int i = startix; i < endix; i++){
								float4 sphere = spheres[i];
								float sq_radius = sphere.w * sphere.w;
								float3 p_a = sphere.xyz - origin.xyz;
								float p_an = dot(p_a, ddir.xyz);
								float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
								if(sq_ndist <= sq_radius){
									if ((p_an - sphere.w) < rtmp.closest_distance){
										float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
										if (dist_exact > 0.0f && dist_exact < rtmp.closest_distance){
											rtmp.state = change = 1;
											rtmp.closest_distance = dist_exact;
											rtmp.sphere_hit = sphere;//.xyz;
											//rtmp.sphere_color = (float4)(0.0f,1.0f,0.0f,0.0f);
											rtmp.sphere_color = sphere_colors[i];
											id = i;
										}
									}
								}
							}
						}
					}
				}
			}
        }
		if(change){
			//if(id >= 0) rtmp.sphere_color = sphere_colors[id];
			d[x] = rtmp;
		}
    }
}


/*
#define GRPSIZE 256

kernel void distmap(global float4* restrict spheres, 
                    global float4* restrict bounding_spheres, 
                    global int* restrict bounding_ref, 
                    global float4* restrict ray_origin,
                    float4 big_bound,
                    int bank,
                    int bounding_start,
                    int bounding_end,
                    int start,
                    int end, 
                    global float4* restrict dir, 
                    //global fi4* td,
                    global fi4* d)
{
    uint x = get_global_id(0);
    fi4 rtmp = d[x];
    float4 ddir = dir[x];
    float4 origin = ray_origin[x];
    if(bank == 0 && start == 0){
        rtmp.x = 999.0f;
        rtmp.id = -2;
    }
    float sq_bbradius = big_bound.w * big_bound.w;
    //float3 bbp_a = big_bound.xyz - origin.xyz;
    float bbp_an = dot(big_bound.xyz - origin.xyz, ddir.xyz);
    float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));

	uint locid = get_local_id(0);
	int grpcount = (bounding_end - bounding_start) / GRPSIZE + 1;
	local float4 loc_bounding_spheres[GRPSIZE];
	local int loc_bounding_ref[GRPSIZE];
	local int bref0;
	bref0 = 0;

	for(int bgrp = 0; bgrp < grpcount; bgrp++){
		int bstart = bounding_start + bgrp*GRPSIZE;
		barrier(CLK_LOCAL_MEM_FENCE);
		if(locid == 0 && bstart > 0) bref0 = bounding_ref[bstart-1];
		loc_bounding_spheres[locid] = bounding_spheres[bstart + locid];
		barrier(CLK_LOCAL_MEM_FENCE);
		loc_bounding_ref[locid] = bounding_ref[bstart + locid];
		barrier(CLK_LOCAL_MEM_FENCE);

		if((rtmp.id != -3) && (sq_bbndist <= sq_bbradius)){
			for(int bitem = 0; bitem < GRPSIZE; bitem++){
				int b = bstart + bitem;
				if(b <= bounding_end){
					float4 bsphere = loc_bounding_spheres[bitem];
					float sq_bradius = bsphere.w * bsphere.w;
					//float3 bp_a = bsphere.xyz - origin.xyz;
					float bp_an = dot(bsphere.xyz - origin.xyz, ddir.xyz);
					float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
					if(sq_bndist <= sq_bradius){
						if((bp_an - bsphere.w) < rtmp.x){
							float bdist_exact = bp_an - native_sqrt(sq_bradius - sq_bndist);
							if(bdist_exact < rtmp.x){
								//int startix = (b == 0) ? 0 : ((bitem == 0) ? bref0 : loc_bounding_ref[bitem-1]);
								int startix = (bitem == 0) ? bref0 : loc_bounding_ref[bitem-1];
								startix = max(start,startix);
								int endix = min(end, loc_bounding_ref[bitem]);

								for(int i = startix; i < endix; i++){
									float4 sphere = spheres[i];
									float sq_radius = sphere.w * sphere.w;
									//float3 p_a = sphere.xyz - origin.xyz;
									float p_an = dot(sphere.xyz - origin.xyz, ddir.xyz);
									float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
									if(sq_ndist <= sq_radius){
										if((p_an - sphere.w) < rtmp.x){
											float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
											if(dist_exact > 0.0f && dist_exact < rtmp.x){
												rtmp.x = dist_exact;
												rtmp.id = i;
											}
										}
									}
								}
							}
						}
					}
				}
			}
        }
    }
    d[x] = rtmp;
}
*/

kernel void shadow(global float4* restrict spheres, 
                   global float4* restrict bounding_spheres, 
                   global unsigned int* restrict bounding_ref, 
                   float4 big_bound,
                   int bounding_start,
                   int bounding_end,
                   unsigned int start,
                   unsigned int end, 
				   float4 lightsource,
                   //global fi4* td,
                   global RaytraceInfo* restrict d)
{
    int x = get_global_id(0);
    //RaytraceInfo rtmp = d[x];
    if((d[x].state == 1) && (d[x].shadow != 0)){
		float4 origin = d[x].sphere_hit;
		float3 dir = normalize(lightsource.xyz - origin.xyz);
		origin.xyz += 1.0001f * dir.xyz * origin.w;
        float sq_bbradius = big_bound.w * big_bound.w;
        float3 bbp_a = big_bound.xyz - origin.xyz;
        float bbp_an = dot(bbp_a, dir);
        float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * dir));
        if(sq_bbndist > sq_bbradius){
            return;
        }
		float dist_origin2light = native_sqrt(sq_point(lightsource.xyz, origin.xyz));
        for(int b = bounding_start; b <= bounding_end; b++){
            float4 bsphere = bounding_spheres[b];
            float sq_bradius = bsphere.w * bsphere.w;
            float3 bp_a = bsphere.xyz - origin.xyz;
            float bp_an = dot(bp_a, dir);
            float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * dir));
            if(sq_bndist <= sq_bradius){
                unsigned int startix = (b == 0) ? 0 : bounding_ref[b-1];
                startix = max(start,startix);
                unsigned int endix = min(end, bounding_ref[b]);
                for(unsigned int i = startix; i < endix; i++){
                    float4 sphere = spheres[i];
                    float sq_radius = sphere.w * sphere.w;
                    float3 p_a = sphere.xyz - origin.xyz;
                    float p_an = dot(p_a, dir);
					if(p_an <= dist_origin2light){
						float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * dir));
						if(sq_ndist <= sq_radius){
							float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
							if(dist_exact >= 0.0f){
								d[x].shadow = 0;
								return;
							}
						}
					}
                }
            }
        }
    }
}
/*
kernel void shadow2(global float4* restrict spheres, 
                   global float4* restrict bounding_spheres, 
                   global int* restrict bounding_ref, 
                   float4 big_bound,
                   int bounding_start,
                   int bounding_end,
                   int start,
                   int end, 
				   global float4* restrict lightsource,
				   int lightsource_count,
                   //global fi4* td,
                   global RaytraceInfo* restrict d)
{
    int x = get_global_id(0);
    //RaytraceInfo rtmp = d[x];
	unsigned int oldshadow = d[x].shadow;  //0=in light, 1=in shadow
	unsigned int ls_test = ((1 << lightsource_count) - 1) & (~oldshadow);  //the lightsources we still need to test (0=skip, 1=test)
	if((d[x].state == 1) && (ls_test != 0)){
		float4 origin = d[x].sphere_hit;
		unsigned int newshadow = oldshadow;

		float sq_bbradius = big_bound.w * big_bound.w;
		float3 bbp_a = big_bound.xyz - origin.xyz;
		float3 dir[32];
		int original_lightsource_index[32];
		int current_lightsource_count = 0;
		for(int ls = 0; ls < lightsource_count; ls++){
			if(ls_test & (1 << ls)) {
				dir[current_lightsource_count] = normalize(lightsource[ls].xyz - origin.xyz);
				float bbp_an = dot(bbp_a, dir[current_lightsource_count]);
				float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * dir[ls]));
				if(sq_bbndist <= sq_bbradius){
					original_lightsource_index[current_lightsource_count] = ls;
					current_lightsource_count++;
				}
			}
		}
		if(current_lightsource_count == 0) return;
		float dist_origin2light[32];
		for(int ls = 0; ls < current_lightsource_count; ls++){
			dist_origin2light[ls] = native_sqrt(sq_point(lightsource[original_lightsource_index[ls]].xyz, origin.xyz));
		}
		int bound_lightsource_index[32];
		for(int b = bounding_start; b <= bounding_end; b++){
			float4 bsphere = bounding_spheres[b];
			float sq_bradius = bsphere.w * bsphere.w;
			float3 bp_a = bsphere.xyz - origin.xyz;
			int bound_lightsource_count = 0;
			for(int ls = 0; ls < current_lightsource_count; ls++) {
				float bp_an = dot(bp_a, dir[ls]);
				float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * dir[ls]));
				if(sq_bndist <= sq_bradius) {
					bound_lightsource_index[bound_lightsource_count++] = ls;
				}
			}
			if(bound_lightsource_count){
				int startix = (b == 0) ? 0 : bounding_ref[b-1];
				startix = max(start,startix);
				int endix = min(end, bounding_ref[b]);
				for(int i = startix; i < endix; i++){
					float4 sphere = spheres[i];
					float sq_radius = sphere.w * sphere.w;
					float3 p_a = sphere.xyz - origin.xyz;
					for(int ls = 0; ls < bound_lightsource_count; ls++) {
						int current_index = bound_lightsource_index[ls];
						float p_an = dot(p_a, dir[current_index]);
						if(p_an <= dist_origin2light[current_index]){
							float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * dir[current_index]));
							if(sq_ndist <= sq_radius){
								float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
								if(dist_exact >= 0.0f){
									newshadow |= (1 << original_lightsource_index[current_index]); //pixel is shadowed
									//recalculate indexes to be able to loop based on number of lightsources
									for(int lst = current_index + 1; lst < current_lightsource_count; lst++) {
										original_lightsource_index[lst] = original_lightsource_index[lst - 1];
										dir[lst] = dir[lst - 1];
										dist_origin2light[lst] = dist_origin2light[lst - 1];
									}
									for(int lst = ls + 1; lst < bound_lightsource_count; lst++) {
										bound_lightsource_index[lst] = bound_lightsource_index[lst - 1];
									}
									current_lightsource_count--;
									bound_lightsource_count--;
									ls--;
								}
							}
						}
					}
				}
			}
		}
		if(newshadow != oldshadow) d[x].shadow = newshadow;
	}
}
*/
kernel void ghostmap(global float4* spheres, 
                    global float4* bounding_spheres, 
                    global int* bounding_ref, 
                    global float* colors, 
                    float4 origin,
                    float4 big_bound,
                    int bank,
                    int bounding_start,
                    int bounding_end,
                    int start,
                    int end, 
                    global float4* dir, 
                    global float4* colorIn,
                    global float4* colorOut)
{
    int x = get_global_id(0);
    float4 ctmp = colorIn[x];
    float4 ddir = dir[x];
    if(bank == 0 && start == 0){
        ctmp = (float4)(0.0f,0.0f,0.0f,0.0f);
    }
    float sq_bbradius = big_bound.w * big_bound.w;
    float3 bbp_a = big_bound.xyz - origin.xyz;
    float bbp_an = dot(bbp_a, ddir.xyz);
    float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));
    if(sq_bbndist <= sq_bbradius){
		for(int b = bounding_start; b <= bounding_end; b++){
			float4 bsphere = bounding_spheres[b];
			float sq_bradius = bsphere.w * bsphere.w;
			float3 bp_a = bsphere.xyz - origin.xyz;
			float bp_an = dot(bp_a, ddir.xyz);
			float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
			if(sq_bndist <= sq_bradius){
				int startix = (b == 0) ? 0 : bounding_ref[b-1];
				startix = max(start,startix);
				int endix = min(end, bounding_ref[b]);
				for(int i = startix; i < endix; i++){
					float4 sphere = spheres[i];
					float sq_radius = sphere.w * sphere.w;
					float3 p_a = sphere.xyz - origin.xyz;
					float p_an = dot(p_a, ddir.xyz);
					float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
					if(sq_ndist < sq_radius){
						float ff = 1.0 - (sq_ndist / sq_radius);
						//ctmp += ff * (float4)(colors[i].xyz, 1.0);
						//float3[x] gets aligned as float4[x]
						ctmp.x += ff * colors[4*i];
						ctmp.y += ff * colors[4*i+1];
						ctmp.z += ff * colors[4*i+2];
						ctmp.w += ff;
					}
				}
			}
		}
	}
    colorOut[x] = ctmp;
}
