//The contents of this file is public domain. No warranty implied. Use at your own risk.
typedef struct tag_RaytraceInfo
{
    int state;
	float closest_distance, mirror_factor;
	unsigned int shadow;
	float4 sphere_hit;
	float4 sphere_color;
} RaytraceInfo;

typedef struct tag_ShadowInfo
{
	float3 position;
	unsigned int shadow;
} ShadowInfo;

float sq_point(float3 p, float3 c)
{
    return ((p.x - c.x) * (p.x - c.x)) + ((p.y - c.y) * (p.y - c.y)) + ((p.z - c.z) * (p.z - c.z));
}

float3 reflect(float3 I, float3 N)
{
	return I - N * dot(N, I) * 2.0f;
}

/*
kernel void distmap(global half* restrict spheres, 
                    global half* restrict bounding_spheres, 
                    global unsigned int* restrict bounding_ref, 
					global half* restrict sphere_colors,
                    global float4* restrict ray_origin,
                    float4 big_bound,
                    int bounding_start,
                    int bounding_end,
                    unsigned int start,
                    unsigned int end, 
                    global float4* restrict dir, 
                    global RaytraceInfo* restrict d)
{
    int x = get_global_id(0);
    RaytraceInfo rtmp = d[x];
    if(rtmp.state != 2){
		int id = -1;
		int change = 0;
        float4 ddir = dir[x];
        float4 origin = ray_origin[x];
        float sq_bbradius = big_bound.w * big_bound.w;
        float3 bbp_a = big_bound.xyz - origin.xyz;
        float bbp_an = dot(bbp_a, ddir.xyz);
        float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));
        if(sq_bbndist <= sq_bbradius){
			for(int b = bounding_start; b <= bounding_end; b++){
				float4 bsphere = vloada_half4(b, bounding_spheres);
				float sq_bradius = bsphere.w * bsphere.w;
				float3 bp_a = bsphere.xyz - origin.xyz;
				float bp_an = dot(bp_a, ddir.xyz);
				float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
				if(sq_bndist <= sq_bradius){
					if((bp_an - bsphere.w) < rtmp.x){
						float bdist_exact = bp_an - native_sqrt(sq_bradius - sq_bndist);
						if(bdist_exact < rtmp.x){
							unsigned int startix = (b == 0) ? 0 : bounding_ref[b-1];
							startix = max(start,startix);
							unsigned int endix = min(end, bounding_ref[b]);
							for(unsigned int i = startix; i < endix; i++){
								float4 sphere = vloada_half4(i, spheres);
								sphere.xyz += bsphere.xyz;
								float sq_radius = sphere.w * sphere.w;
								float3 p_a = sphere.xyz - origin.xyz;
								float p_an = dot(p_a, ddir.xyz);
								float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
								if(sq_ndist <= sq_radius){
									if((p_an - sphere.w) < rtmp.x){
										float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
										if(dist_exact > 0.0f && dist_exact < rtmp.x){
											rtmp.state = change = 1;
											rtmp.x = dist_exact;
											rtmp.sphere_hit = sphere;//.xyz;
											id = i;
										}
									}
								}
							}
						}
					}
				}
			}
			if(change){
				if(id >= 0) rtmp.sphere_color = vloada_half4(id, sphere_colors);
				d[x] = rtmp;
			}
        } //if(sq_bbndist <= sq_bbradius)
    } //if(rtmp.state != 2)
}
*/

/*
#define GRPSIZE 256

kernel void distmap(global float4* restrict spheres, 
                    global float4* restrict bounding_spheres, 
                    global int* restrict bounding_ref, 
                    global float4* restrict ray_origin,
                    float4 big_bound,
                    int bank,
                    int bounding_start,
                    int bounding_end,
                    int start,
                    int end, 
                    global float4* restrict dir, 
                    //global fi4* td,
                    global fi4* d)
{
    uint x = get_global_id(0);
    fi4 rtmp = d[x];
    float4 ddir = dir[x];
    float4 origin = ray_origin[x];
    if(bank == 0 && start == 0){
        rtmp.x = 999.0f;
        rtmp.id = -2;
    }
    float sq_bbradius = big_bound.w * big_bound.w;
    //float3 bbp_a = big_bound.xyz - origin.xyz;
    float bbp_an = dot(big_bound.xyz - origin.xyz, ddir.xyz);
    float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));

	uint locid = get_local_id(0);
	int grpcount = (bounding_end - bounding_start) / GRPSIZE + 1;
	local float4 loc_bounding_spheres[GRPSIZE];
	local int loc_bounding_ref[GRPSIZE];
	local int bref0;
	bref0 = 0;

	for(int bgrp = 0; bgrp < grpcount; bgrp++){
		int bstart = bounding_start + bgrp*GRPSIZE;
		barrier(CLK_LOCAL_MEM_FENCE);
		if(locid == 0 && bstart > 0) bref0 = bounding_ref[bstart-1];
		loc_bounding_spheres[locid] = bounding_spheres[bstart + locid];
		barrier(CLK_LOCAL_MEM_FENCE);
		loc_bounding_ref[locid] = bounding_ref[bstart + locid];
		barrier(CLK_LOCAL_MEM_FENCE);

		if((rtmp.id != -3) && (sq_bbndist <= sq_bbradius)){
			for(int bitem = 0; bitem < GRPSIZE; bitem++){
				int b = bstart + bitem;
				if(b <= bounding_end){
					float4 bsphere = loc_bounding_spheres[bitem];
					float sq_bradius = bsphere.w * bsphere.w;
					//float3 bp_a = bsphere.xyz - origin.xyz;
					float bp_an = dot(bsphere.xyz - origin.xyz, ddir.xyz);
					float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
					if(sq_bndist <= sq_bradius){
						if((bp_an - bsphere.w) < rtmp.x){
							float bdist_exact = bp_an - native_sqrt(sq_bradius - sq_bndist);
							if(bdist_exact < rtmp.x){
								//int startix = (b == 0) ? 0 : ((bitem == 0) ? bref0 : loc_bounding_ref[bitem-1]);
								int startix = (bitem == 0) ? bref0 : loc_bounding_ref[bitem-1];
								startix = max(start,startix);
								int endix = min(end, loc_bounding_ref[bitem]);

								for(int i = startix; i < endix; i++){
									float4 sphere = spheres[i];
									float sq_radius = sphere.w * sphere.w;
									//float3 p_a = sphere.xyz - origin.xyz;
									float p_an = dot(sphere.xyz - origin.xyz, ddir.xyz);
									float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
									if(sq_ndist <= sq_radius){
										if((p_an - sphere.w) < rtmp.x){
											float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
											if(dist_exact > 0.0f && dist_exact < rtmp.x){
												rtmp.x = dist_exact;
												rtmp.id = i;
											}
										}
									}
								}
							}
						}
					}
				}
			}
        }
    }
    d[x] = rtmp;
}
*/
/*
kernel void shadow(global half* restrict spheres, 
                   global half* restrict bounding_spheres, 
                   global unsigned int* restrict bounding_ref, 
                   float4 big_bound,
                   int bounding_start,
                   int bounding_end,
                   unsigned int start,
                   unsigned int end, 
				   float4 lightsource,
                   global float4* restrict d)
{
    int x = get_global_id(0);
	float4 origin = d[x];
	if(origin.w > 0.0f){
		float3 dir = normalize(lightsource.xyz - origin.xyz);
		origin.xyz += (1.0001f * origin.w) * dir.xyz;
		float sq_bbradius = big_bound.w * big_bound.w;
		float3 bbp_a = big_bound.xyz - origin.xyz;
		float bbp_an = dot(bbp_a, dir);
		float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * dir));
		if(sq_bbndist <= sq_bbradius){
			float dist_origin2light = native_sqrt(sq_point(lightsource.xyz, origin.xyz));
			for(int b = bounding_start; b <= bounding_end; b++){
				float4 bsphere = vloada_half4(b, bounding_spheres);
				float sq_bradius = bsphere.w * bsphere.w;
				float3 bp_a = bsphere.xyz - origin.xyz;
				float bp_an = dot(bp_a, dir);
				float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * dir));
				if(sq_bndist <= sq_bradius){
					unsigned int startix = (b == 0) ? 0 : bounding_ref[b-1];
					startix = max(start,startix);
					unsigned int endix = min(end, bounding_ref[b]);
					for(unsigned int i = startix; i < endix; i++){
						float4 sphere = vloada_half4(i, spheres);
						sphere.xyz += bsphere.xyz;
						float sq_radius = sphere.w * sphere.w;
						float3 p_a = sphere.xyz - origin.xyz;
						float p_an = dot(p_a, dir);
						if(p_an <= dist_origin2light){
							float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * dir));
							if(sq_ndist <= sq_radius){
								float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
								if(dist_exact >= 0.0f){
									d[x].w = -origin.w;
									return;
								}
							}
						}
					}
				}
			}
		}
	}
}
*/
/*
kernel void shadow(global half* restrict spheres, 
                   global half* restrict bounding_spheres, 
                   global unsigned int* restrict bounding_ref, 
                   float4 big_bound,
                   int bounding_start,
                   int bounding_end,
                   unsigned int start,
                   unsigned int end, 
				   float4 lightsource,
                   global RaytraceInfo* restrict d)
{
    int x = get_global_id(0);
    //RaytraceInfo rtmp = d[x];
    if((d[x].state == 1) && (d[x].shadow != 0)){
		float4 origin = d[x].sphere_hit;
		float3 dir = normalize(lightsource.xyz - origin.xyz);
		origin.xyz += (1.0001f * origin.w) * dir.xyz;
        float sq_bbradius = big_bound.w * big_bound.w;
        float3 bbp_a = big_bound.xyz - origin.xyz;
        float bbp_an = dot(bbp_a, dir);
        float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * dir));
        if(sq_bbndist <= sq_bbradius){
			float dist_origin2light = native_sqrt(sq_point(lightsource.xyz, origin.xyz));
			for(int b = bounding_start; b <= bounding_end; b++){
				float4 bsphere = vloada_half4(b, bounding_spheres);
				float sq_bradius = bsphere.w * bsphere.w;
				float3 bp_a = bsphere.xyz - origin.xyz;
				float bp_an = dot(bp_a, dir);
				float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * dir));
				if(sq_bndist <= sq_bradius){
					unsigned int startix = (b == 0) ? 0 : bounding_ref[b-1];
					startix = max(start,startix);
					unsigned int endix = min(end, bounding_ref[b]);
					for(unsigned int i = startix; i < endix; i++){
						float4 sphere = vloada_half4(i, spheres);
						sphere.xyz += bsphere.xyz;
						float sq_radius = sphere.w * sphere.w;
						float3 p_a = sphere.xyz - origin.xyz;
						float p_an = dot(p_a, dir);
						if(p_an <= dist_origin2light){
							float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * dir));
							if(sq_ndist <= sq_radius){
								float dist_exact = p_an - native_sqrt(sq_radius - sq_ndist);
								if(dist_exact >= 0.0f){
									d[x].shadow = 0;
									return;
								}
							}
						}
					}
		        }
            }
        }
    }
}
*/
kernel void ghostmap(global half* spheres, 
                    global half* bounding_spheres, 
                    global int* bounding_ref, 
                    global float* colors, 
                    float4 origin,
                    float4 big_bound,
                    int bank,
                    int bounding_start,
                    int bounding_end,
                    int start,
                    int end, 
                    global float4* dir, 
                    global float4* colorIn,
                    global float4* colorOut)
{
    int x = get_global_id(0);
    float4 ctmp = colorIn[x];
    float4 ddir = dir[x];
    if(bank == 0 && start == 0){
        ctmp = (float4)(0.0f,0.0f,0.0f,0.0f);
    }
    float sq_bbradius = big_bound.w * big_bound.w;
    float3 bbp_a = big_bound.xyz - origin.xyz;
    float bbp_an = dot(bbp_a, ddir.xyz);
    float sq_bbndist = sq_point(big_bound.xyz, origin.xyz + (bbp_an * ddir.xyz));
    if(sq_bbndist <= sq_bbradius){
		for(int b = bounding_start; b <= bounding_end; b++){
			float4 bsphere = vloada_half4(b, bounding_spheres); //float4 bsphere = bounding_spheres[b];
			float sq_bradius = bsphere.w * bsphere.w;
			float3 bp_a = bsphere.xyz - origin.xyz;
			float bp_an = dot(bp_a, ddir.xyz);
			float sq_bndist = sq_point(bsphere.xyz, origin.xyz + (bp_an * ddir.xyz));
			if(sq_bndist <= sq_bradius){
				int startix = (b == 0) ? 0 : bounding_ref[b-1];
				startix = max(start,startix);
				int endix = min(end, bounding_ref[b]);
				for(int i = startix; i < endix; i++){
					float4 sphere = vloada_half4(i, spheres);	//float4 sphere = spheres[i];
					float sq_radius = sphere.w * sphere.w;
					float3 p_a = sphere.xyz - origin.xyz;
					float p_an = dot(p_a, ddir.xyz);
					float sq_ndist = sq_point(sphere.xyz, origin.xyz + (p_an * ddir.xyz));
					if(sq_ndist < sq_radius){
						float ff = 1.0 - (sq_ndist / sq_radius);
						//ctmp += ff * (float4)(colors[i].xyz, 1.0);
						//float3[x] gets aligned as float4[x]
						ctmp.x += ff * colors[4*i];
						ctmp.y += ff * colors[4*i+1];
						ctmp.z += ff * colors[4*i+2];
						ctmp.w += ff;
					}
				}
			}
		}
	}
    colorOut[x] = ctmp;
}
